/*
 * $Id: optproc.C,v 1.33 1995/10/05 21:09:51 asf Exp $
 */

/*
    This file is part of Addventure
    Copyright (C) 1994, 1995  Allen S. Firstenberg
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 * Process the option entry screen:
 * 1)  Ensure that the temporary file exists, the target rooms are all
 *     valid, and ensure that each option description is filled in, and
 *     that the room was corectly started.
 * 2)  Determine which new rooms need to be created, and update the
 *     game "next room" counter.
 * 3)  Make the "html" and "lock" files for new rooms.
 * 4)  Output the links, signature, and date to the temp file.
 * 5)  Overwrite the old html file with the temp file.
 * 6)  Add to the list of target rooms, if appropriate
 * 7)  Update the statistics
 * 7a) If the author of the parent wants, notify them that a child has
 *     been written.
 * 7b) If the child's author wants, but the children in the notify list.
 * 8)  Remove the indicator that the room was started (opened for editing)
 * 9)  Return a "ready for entry" page
 */

#include <time.h>

#include <post-query.H>

#include <htmllist.H>
#include <htmlfile.H>
#include <htmltext.H>

#include "addvutil.H"
#include "backlink.H"
#include "makeroom.H"
#include "lockfile.H"
#include "openfile.H"
#include "nextroom.H"
#include "setaddv.H"
#include "error.H"
#include "option.H"
#include "stat.H"
#include "response.H"

/***************************************************************************/

void extract_info( int numopt, BacklinkType linktype,   // In
                   PostQuery *query, BackLink *rooms,   // In
                   int allow_anchors,                   // In
                   OptionArray &option, int &newopt,    // to be filled in
                   int &links_bad, int &options_bad ){  // to be returned
  /*
   * A seriously overloaded function that accomplishes the following (not
   * necessarily in this order):
   *
   * 1) Convert all the requested room numbers and place them
   *    into an created array.
   * 2) Ensure each room number is legal.  Set "links_bad" if they are to
   *    0 and to 1 if not.
   * 3) Determine how many new rooms need to be generated
   * 4) Make a pointer from the array to the description of each
   *    option
   * 5) Ensure each option description is legal.  Set "options_bad"
   *    to 0 if they are and to 1 if its empty or 2 if it has bad HTML.
   */

  int  co;          // Counter
  char field[20];   // What field we are trying to get from the query
  int  room;       
  char *room_str;
  char *desc_str;

  // Assume everything is ok
  links_bad = 0;
  options_bad = 0;

  // Start out with no new rooms
  newopt = 0;

  for( co=1; co<=numopt; co++ ){

    // Get the option description
    sprintf( field, "option%d", co );
    desc_str = query->lookup( field );

    // Make sure it has something in it
    if( !null_strlen( desc_str ) ){
      // No string, or the string is empty
      options_bad = 1;
    }

    // Make sure there is no anchoring in here
    if( (!allow_anchors) && (!options_bad) && has_evil_html( desc_str ) ){
      // There is an anchor.  Bad news.
      options_bad = 2;
    }

    // Get the room to link to
    switch( linktype ){
    case BL_NONE:
      // Nothing was entered.  So it must be created
      room = 0;
      break;
    case BL_NAME:
    case BL_NUMBER:
      // These were the sorted list
      sprintf( field, "link%d", co );
      room_str = query->lookup( field );
      room = backlink_to_room( room_str );
      break;
    case BL_ENTRY:
      // The number itself should have been entered
      sprintf( field, "link%d", co );
      room_str = query->lookup( field );
      room = atoi( room_str );
      break;
    }

    // Make sure the link is legal
    if( room ){
      // There was a room specified
      if( !rooms->room_exists( room ) ){
        // The room is not in the list
        links_bad = 1;
      }
    }

    // Determine if this is a new rooms
    if( room == 0 ){
      // It is a new room
      newopt++;
    }

    // Assign it to the array
    (option[co])->set( desc_str, room );
  }
  
}

/***************************************************************************/

HtmlNode *make_temp2_option( char *urlpath, int room, char *desc ){
  /*
   * Return a node suitable for including in an HtmlList that contains
   * a link to the new room.
   */

  char *nodeline;

  char *pathroom = make_fname( NULL, room, "html" );
  char *url = make_url( NULL, NULL, pathroom );

  HtmlText t;
  t.link( url, desc );

  nodeline = (char *)malloc( null_strlen(t.val()) );
  strcpy( nodeline, t.val() );

  HtmlNode *ret = new HtmlNode( nodeline );

  free( url );
  free( pathroom );

  return ret;
}

//---------------------------------------------------------------------------

int make_temp2( char *urlpath, char *filename,
                int numopts, OptionArray &options,
                char *sig, int parent,
                char *roomword, int backlink, int parentlink ){
  /*
   * Create the second part of the room
   * Return 1 if success and 0 if problems
   *
   * "backlink" indicates if this room can be linked to
   * "parentlink" indicates if we should link to the parent
   */
  int ret;

  FILE *room;

  char *file;
  char *url;

  int co;

  // Create the list of links
  HtmlList optlist;
  HtmlNode *node;
  
  for( co=1; co<=numopts; co++ ){
    node = make_temp2_option( urlpath,
                              options[co]->room(),
                              options[co]->description() );
    optlist.append( *node );
  }

  set_permissions();
  if( room = fopen( filename, "a" ) ){
    // File opened

    HtmlFile f( room );

    // Put the list of the choices in
    f.numlist( &optlist );

    // Sometimes we can link to the parent
    if( parentlink ){
      file = make_fname( NULL, parent, "html" );
      url = make_url( NULL, NULL, file );
      f.link( url, "Go back" );
      f.puts( " - Go to the parent " );
      f.puts( roomword );
      f.puts( ".\n" );
      free( url );
      free( file );
    }

    // Add the signature if it exists
    f.line();
    if( sig ){
      // It was signed
      f.address( sig );  f.para();
    }

    // Add the date
    time_t t = time(NULL);
    f.puts( ctime( &t ) );  f.para();

    // Put up a notice if we can link to this room
    if( backlink ){
      f.puts( "<i>Linking Enabled</i>" );  f.para();
    }

    fclose( room );
    ret = 1;
  } else {
    // Open failed
    ret = 0;
  }

  return ret;
}

/***************************************************************************/

void ready_for_entry( char *urlpath, int room, int parent,
                      char *title, char *roomword ){
  char pagetitle[160];
  char *file;
  char *url;

  // Start the page
  HtmlFile f(stdout);
  f.content_type();

  // Generate the title
  sprintf( pagetitle, "%s %d ready for entry", roomword, room );
  f.title( pagetitle );

  // Generate the header
  f.header( 1, "Ready for entry" );
  f.header( 2, title );

  // Create the link to the room
  file = make_fname( urlpath, room, "html" );
  url = make_url( NULL, NULL, file, "" );      // Need to append a fake query
  f.link( url, "Enter!" );
  f.puts( " - Go to new " );  f.puts( roomword );  f.para();
  free( url );
  free( file );

  // Create the link to the parent, just in case
  file = make_fname( urlpath, parent, "html" );
  url = make_url( NULL, NULL, file );
  f.link( url, "Go back" );
  f.puts( " - Go to the " );
  f.puts( roomword );
  f.puts( " you were just in.\n" );
  free( url );
  free( file );
}

/***************************************************************************/

#define MAXENTRIES 100

main(){
  PostQuery query( MAXENTRIES );

  int room;
  int parent;
  int numopt;      // The number of options
  char *title;     // The title of the room
  BacklinkType backlink;  // The backlink type value
  
  char *urlpath;   // The path relative to the server root
  char *filepath;  // The absolute path in the file system
  char *binpath;   // Where the cgi scrpts are

  char *host;      // Where they're connecting from

  char *tempfile;  // The temporary file storing the new room
  char *htmlfile;  // The permanent name fo the file

  int invalid_link;
  int invalid_options;
  int newopt;            // How many options require new rooms

  char *signature;       // If the person signed it
  int  linkhere;         // If this room can be linked to later
  char *linkherestr;
  int  sendmail;         // Should we send mail when a child is created
  char *sendmailstr;
  char *roomword;        // What do we call a "room"
  char *email;           // The email address
  int  parentlink;       // Should we link to the parent?
  int  goback;           // The value from the setting file

  int emptyroom;         // The first new room we can create

  int co;

  if( !query.valid() ){
    // Something is wrong with the query
    addventure_error( OPTPROC, ADDVERR_UNKNOWN );
  } else {
    // Process the data

    // Get the hidden data
    room     = atoi( query.lookup( "room" ) );
    parent   = atoi( query.lookup( "parent" ) );
    numopt   = atoi( query.lookup( "numoptions" ) );
    title    = query.lookup( "title" );
    backlink = get_backlink_type( query.lookup( "backlink" ) );

    // We don't know the array size till now
    OptionArray optionlist(numopt);   // The links and descriptions

    // Get the data passed by environment variables
    urlpath  = getenv( "PATH_INFO" );
    filepath = getenv( "PATH_TRANSLATED" );
    binpath  = get_path( getenv( "SCRIPT_NAME" ) );
    host     = getenv( "REMOTE_HOST" );

    SetAddv setting( filepath );    // Read from the game values
    roomword = setting.roomword();

    // Get the list of legal rooms
    BackLink rooms( filepath, roomword );

    // Load all of the option data
    int allow_anchors = setting.evil_html();
    extract_info( numopt, backlink, &query, &rooms, allow_anchors,
                  optionlist, newopt, invalid_link, invalid_options );

    sendmailstr = query.lookup( "sendmail" );
    if( (sendmailstr) && (!strcmp( sendmailstr, "1" )) ){
      // It was set

      email = query.lookup( "email" );
      if( (email) && (strcmp( email, "" )) ){
        // The address was set
        sendmail = 1;
      } else {
        // No address set
        sendmail = 0;
      }
      
    } else {
      // It wasn't set
      sendmail = 0;
    }

    linkherestr = query.lookup( "linkhere" );
    if( (linkherestr) && (!strcmp( linkherestr, "1" )) ){
      // It was set
      linkhere = 1;
    } else {
      // It wasn't set
      linkhere = 0;
    }

    tempfile = make_fname( filepath, room, "temp" );

    // Where do we store the open data
    OpenFile openfile( filepath, room );

    if( !openfile.exists() ){
      // This wasn't approached normally
      addventure_error( OPTPROC, ADDVERR_NOT_EXIST, roomword );
    } else if( !file_exists( tempfile ) ){
      // Oops!  The temporary file doesn't yet exist
      addventure_error( OPTPROC, ADDVERR_NOT_EXIST, roomword );
    } else if( invalid_link ){
      // It was an entry type, and the rooms weren't in the list
      addventure_error( OPTPROC, ADDVERR_INVALID_TARGET, roomword );
    } else if( invalid_options == 1 ){
      // There were options that were empty
      addventure_error( OPTPROC, ADDVERR_EMPTY, roomword );
    } else if( invalid_options == 2 ){
      // There were options that had nothing legal in them
      addventure_error( OPTPROC, ADDVERR_HTML, roomword );
    } else {
      // Hey!  We can do something!
      int room_count_ok = 1;

      if( newopt ){
        // Get the first room we can use and update the "new room" counter
        // @@@ Warning!  There is a race condition here, so it MUST be short
        NextRoom nextroom( filepath );
        emptyroom = nextroom;

        // @@@ This voilates teh concept of "short", but I'm getting too
        //     many problems otherwise
        if( emptyroom > 0 ){
          // We were able to get the next room without error
          nextroom = (emptyroom + newopt);

          // Create the new rooms and locks (if necessary)
          for( co=1; (co<=numopt) && (room_count_ok); co++ ){
            if( !optionlist[co]->room() ){
              // We will be linking to a new room

              // Determine what the room will be and save it
              optionlist[co]->set( emptyroom );

              // Create the room and lock
              if( make_room( filepath, urlpath, binpath,
                             roomword, emptyroom, room ) ){
                // Created a room
                LockFile lockfile( filepath, emptyroom );
                lockfile.create( room );
              } else {
                // Couldn't create the room
                room_count_ok = 0;
                char error_buffer[200];
                sprintf( error_buffer, "The %s %d already exists.  %s",
                         roomword, emptyroom,
                         "Please make sure we know about the problem."
                       );
                addventure_error( OPTPROC, ADDVERR_NONEXTROOM, roomword,
                                  error_buffer
                                );
                
              }

              // Get ready for the next one
              emptyroom++;
            } // end if not room
          } // end for

        } else {
          // There was an error
          room_count_ok = 0;

          addventure_error( OPTPROC, ADDVERR_NONEXTROOM, roomword,
                            "Please make sure we know about the problem."
                          );
          
        } // end if/else emptyroom > 0
          
      } // end if newopt

      if( room_count_ok ){
        // Get the signature to put onthe room
        signature = query.lookup( "signature" );

        // Determine if we should link back to the parent
        goback = setting.goback();
        switch( goback ){
        case 0:
        case 1:
          // We should either always, or never link to the parent
          parentlink = goback;
          break;
        case -1:
          // We should only link if we are guaranteed there is only
          // going to be one parent (ie - the current room can not
          // be linked to).
          parentlink = !linkhere;
          break;
        }

        // Output the rest of the temp file
        make_temp2( urlpath, tempfile, numopt, optionlist, signature,
                   parent, roomword, linkhere, parentlink );

        // Move the temporary file to the permanent file
        htmlfile = make_fname( filepath, room, "html" );
        rename( tempfile, htmlfile );
        free( tempfile );

        // Remove the record that the file is in progress
        openfile.remove();

        // Add to the list of target rooms if appropriate
        if( linkhere ){
          // Add to the list of target rooms
          rooms.append( title );
        }

        // Update the statistics
        StatFile stats( filepath );
        stats.add( room, optionlist, host, signature );

        // Send mail to the author of the parent, if appropriate
        ResponseFile response( filepath, 1 );
        response.respond( room );
        // Then add these children to the list if the author wants
        if( sendmail ){
          response.add( room, optionlist, email );
        }

        // Return a "ready for entry" room
        ready_for_entry( urlpath, room, parent, title, roomword );

      } // end room_count_ok
      
    } // end if/else/else/else
  }
}


/***************************************************************************/

/*
 * $Log: optproc.C,v $
 * Revision 1.33  1995/10/05  21:09:51  asf
 * Extensive attempts at error checking, trying to make sure it
 * is impossible to overwrite existing rooms.
 *
 * Revision 1.32  1995/05/25  21:01:56  asf
 * The "Go back" link contained the absolute path of the parent link, while
 * none of the children links had this characteristic.  Removed the absolute
 * path.
 *
 * Revision 1.31  1995/02/24  18:07:13  asf
 * Fixed a spacing problem.
 *
 * Revision 1.30  1995/02/12  12:40:16  asf
 * Changes to "correctly" support relative addressing.
 *
 * Revision 1.29  1995/01/28  16:54:34  asf
 * has_anchor() removed and replaced with the more general utility
 * has_bad_html(), as well as the more general settings.
 *
 * Changed wording for the link to the parent.
 *
 * Revision 1.28  1995/01/22  12:39:47  asf
 * Changes to make_temp2_option() to use relative path names when creating
 * the room instead of including the entire absolute path.
 *
 * Revision 1.27  1995/01/01  17:37:28  asf
 * Added the "room word" to the error messages
 *
 * Revision 1.26  1995/01/01  16:31:46  asf
 * Added support for the "room word"
 *
 * Revision 1.25  1994/12/25  11:07:12  asf
 * Added changes to support ADDVERR_HTML.
 *
 * Revision 1.24  1994/12/17  20:14:52  asf
 * Checks the game settings to see if anchors are allowed in options
 * or not.
 *
 * Revision 1.23  1994/12/17  19:10:53  asf
 * Added anchor checking feature (but not the ability to make it an option).
 *
 * Revision 1.22  1994/12/15  21:07:34  asf
 * "sendmail" was probably turned off by a messy computation if "email"
 * was legal.  Fixed (hopefully).
 *
 * Revision 1.21  1994/12/15  20:59:07  asf
 * Added a new box to enter an email address, seperate from the signature.
 *
 * Revision 1.20  1994/12/15  19:47:08  asf
 * The ResponseFile now also stores the parent of each room.
 *
 * Revision 1.19  1994/11/06  13:08:30  asf
 * Added processing to send a response to the author of a room when
 * one of the children is edited.
 *
 * Revision 1.18  1994/10/01  16:14:37  asf
 * Added a paragraph signal after the (optional) address.
 *
 * Revision 1.17  1994/10/01  15:44:42  asf
 * Switched to use of LockFile and OpenFile.
 *
 * Revision 1.16  1994/09/16  15:08:14  asf
 * Several changes that now read the game defaults to determine if a
 * link will be generated to the parent room.
 *
 * Revision 1.15  1994/08/31  18:23:47  asf
 * Link to the parent if we are guaranteed that there is only one parent
 *
 * Revision 1.14  1994/08/31  18:12:07  asf
 * Rooms that can be linked to now display this fact.
 *
 * Revision 1.13  1994/08/30  21:13:59  asf
 * Added calls to the statistics module
 *
 * Revision 1.12  1994/08/26  16:46:21  asf
 * Converted use of the Option structure to the Option class.  Converted use
 * of malloced space for the options to the OptionArray class.  This allowed
 * us to remove some ugly looking code in extract_info()
 *
 * Revision 1.11  1994/08/26  09:08:31  asf
 * Converted to error package.
 *
 * Revision 1.10  1994/08/25  01:09:00  asf
 * Added openfile support.
 *
 * Revision 1.9  1994/08/18  06:37:24  asf
 * Also checked for a NULL value for the linkhere string.
 *
 * Revision 1.8  1994/08/17  23:18:38  asf
 * Fixed a potential bug involving the linkhere variable.
 *
 * Revision 1.7  1994/08/15  19:43:13  asf
 * Converted calls from strlen() to null_strlen() and then optimized some
 * if statements.
 *
 * Revision 1.6  1994/08/14  15:08:04  asf
 * Moved the large if/else/else/else bracketing so that nothing could
 * happen if there had been an error.
 *
 * Revision 1.5  1994/08/14  07:56:11  asf
 * Removed debugging statements
 *
 * Revision 1.4  1994/08/14  00:48:15  asf
 * Added permanent allocation for lines in make_temp2_option().
 * Added more debugging.
 *
 * Revision 1.3  1994/08/14  00:00:59  asf
 * Moved the load for the rooms till _after_ we have a path, which is a good
 * thing.
 *
 * Revision 1.2  1994/08/13  23:25:21  asf
 * Put in many debugging statements
 * Cleaned the memory leak that seemed to be in extract_info()
 *
 * Revision 1.1  1994/08/13  19:11:59  asf
 * Initial revision
 *
 */


